@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2003-2017. All Rights Reserved.
+%% Copyright Ericsson AB 2003-2018. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -20,10 +20,11 @@
 
 -module(zlib).
 
--export([open/0,close/1,deflateInit/1,deflateInit/2,deflateInit/6,
+-export([open/0,close/1,set_controlling_process/2,
+         deflateInit/1,deflateInit/2,deflateInit/6,
          deflateSetDictionary/2,deflateReset/1,deflateParams/3,
          deflate/2,deflate/3,deflateEnd/1,
-         inflateInit/1,inflateInit/2,
+         inflateInit/1,inflateInit/2,inflateInit/3,
          inflateSetDictionary/2,inflateGetDictionary/1, inflateReset/1,
          inflate/2,inflate/3,inflateEnd/1,
          inflateChunk/2,inflateChunk/1,
@@ -68,6 +69,13 @@
 
 -define(MAX_WBITS, 15).
 
+-define(DEFAULT_MEMLEVEL, 8).
+-define(DEFAULT_WBITS, 15).
+
+-define(EOS_BEHAVIOR_ERROR, 0).
+-define(EOS_BEHAVIOR_RESET, 1).
+-define(EOS_BEHAVIOR_CUT, 2).
+
 %% Chunk sizes are hardcoded on account of them screwing with the
 %% predictability of the system. zlib is incapable of trapping so we need to
 %% ensure that it never operates on any significant amount of data.
@@ -79,7 +87,7 @@
 %%------------------------------------------------------------------------
 
 %% Public data types.
--type zstream() :: term().
+-type zstream() :: reference().
 -type zflush() :: 'none' | 'sync' | 'full' | 'finish'.
 
 -type zlevel() ::
@@ -94,11 +102,11 @@
 -type zmethod()     :: 'deflated'.
 
 -record(zlib_opts, {
-        stream :: zstream(),
-        method :: term(),
-        input_chunk_size :: integer(),
-        output_chunk_size :: integer(),
-        flush :: integer()
+        stream :: zstream() | 'undefined',
+        method :: function(),
+        input_chunk_size :: pos_integer(),
+        output_chunk_size :: pos_integer(),
+        flush :: non_neg_integer()
     }).
 
 %%------------------------------------------------------------------------
@@ -121,6 +129,14 @@ close(Z) ->
 close_nif(_Z) ->
     erlang:nif_error(undef).
 
+-spec set_controlling_process(Z, Pid) -> 'ok' when
+      Z :: zstream(),
+      Pid :: pid().
+set_controlling_process(Z, Pid) ->
+    set_controller_nif(Z, Pid).
+set_controller_nif(_Z, _Pid) ->
+    erlang:nif_error(undef).
+
 -spec deflateInit(Z) -> 'ok' when
       Z :: zstream().
 deflateInit(Z) ->
@@ -130,10 +146,7 @@ deflateInit(Z) ->
       Z :: zstream(),
       Level :: zlevel().
 deflateInit(Z, Level) ->
-    deflateInit_nif(Z, arg_level(Level)).
-
-deflateInit_nif(_Z, _Level) ->
-    erlang:nif_error(undef).
+    deflateInit(Z, Level, deflated, ?DEFAULT_WBITS, ?DEFAULT_MEMLEVEL, default).
 
 -spec deflateInit(Z, Level, Method, WindowBits, MemLevel, Strategy) -> 'ok' when
       Z :: zstream(),
@@ -155,7 +168,7 @@ deflateInit_nif(_Z, _Level, _Method, _WindowBits, _MemLevel, _Strategy) ->
 -spec deflateSetDictionary(Z, Dictionary) -> Adler32 when
       Z :: zstream(),
       Dictionary :: iodata(),
-      Adler32 :: integer().
+      Adler32 :: non_neg_integer().
 deflateSetDictionary(Z, Dictionary) ->
     deflateSetDictionary_nif(Z, Dictionary).
 deflateSetDictionary_nif(_Z, _Dictionary) ->
@@ -175,14 +188,13 @@ deflateReset_nif(_Z) ->
 deflateParams(Z, Level0, Strategy0) ->
     Level = arg_level(Level0),
     Strategy = arg_strategy(Strategy0),
+    Progress = deflate(Z, <<>>, sync),
     case deflateParams_nif(Z, Level, Strategy) of
-        buf_error ->
-            %% We had data left in the pipe; flush everything and stash it away
-            %% for the next deflate call before trying again.
-            Output = deflate(Z, <<>>, full),
-            save_progress(Z, deflate, Output),
-            deflateParams_nif(Z, Level, Strategy);
-        Any -> Any
+        ok ->
+            save_progress(Z, deflate, Progress),
+            ok;
+        Other ->
+            Other
     end.
 deflateParams_nif(_Z, _Level, _Strategy) ->
     erlang:nif_error(undef).
@@ -225,16 +237,21 @@ deflateEnd_nif(_Z) ->
 -spec inflateInit(Z) -> 'ok' when
       Z :: zstream().
 inflateInit(Z) ->
-    inflateInit_nif(Z).
-inflateInit_nif(_Z) ->
-    erlang:nif_error(undef).
+    inflateInit(Z, ?DEFAULT_WBITS).
 
 -spec inflateInit(Z, WindowBits) -> 'ok' when
       Z :: zstream(),
       WindowBits :: zwindowbits().
 inflateInit(Z, WindowBits) ->
-    inflateInit_nif(Z, WindowBits).
-inflateInit_nif(_Z, _WindowBits) ->
+    inflateInit(Z, WindowBits, cut).
+
+-spec inflateInit(Z, WindowBits, EoSBehavior) -> 'ok' when
+      Z :: zstream(),
+      WindowBits :: zwindowbits(),
+      EoSBehavior :: error | reset | cut.
+inflateInit(Z, WindowBits, EoSBehavior) ->
+    inflateInit_nif(Z, arg_bitsz(WindowBits), arg_eos_behavior(EoSBehavior)).
+inflateInit_nif(_Z, _WindowBits, _EoSBehavior) ->
     erlang:nif_error(undef).
 
 -spec inflateSetDictionary(Z, Dictionary) -> 'ok' when
@@ -278,7 +295,7 @@ inflate(Z, Data) ->
       Options :: list({exception_on_need_dict, boolean()}),
       Decompressed :: iolist() |
                       {need_dictionary,
-                       Adler32 :: integer(),
+                       Adler32 :: non_neg_integer(),
                        Output :: iolist()}.
 inflate(Z, Data, Options) ->
     enqueue_input(Z, Data),
@@ -340,7 +357,7 @@ exception_on_need_dict(Z, Output) when is_list(Output); is_binary(Output) ->
       Result :: {continue, Output :: iolist()} |
                 {finished, Output :: iolist()} |
                 {need_dictionary,
-                 Adler32 :: integer(),
+                 Adler32 :: non_neg_integer(),
                  Output :: iolist()}.
 safeInflate(Z, Data) ->
     enqueue_input(Z, Data),
@@ -372,7 +389,7 @@ getBufSize_nif(_Z) ->
 
 -spec crc32(Z) -> CRC when
       Z :: zstream(),
-      CRC :: integer().
+      CRC :: non_neg_integer().
 crc32(Z) ->
     crc32_nif(Z).
 crc32_nif(_Z) ->
@@ -381,7 +398,7 @@ crc32_nif(_Z) ->
 -spec crc32(Z, Data) -> CRC when
       Z :: zstream(),
       Data :: iodata(),
-      CRC :: integer().
+      CRC :: non_neg_integer().
 crc32(Z, Data) when is_reference(Z) ->
     erlang:crc32(Data);
 crc32(_Z, _Data) ->
@@ -389,9 +406,9 @@ crc32(_Z, _Data) ->
 
 -spec crc32(Z, PrevCRC, Data) -> CRC when
       Z :: zstream(),
-      PrevCRC :: integer(),
+      PrevCRC :: non_neg_integer(),
       Data :: iodata(),
-      CRC :: integer().
+      CRC :: non_neg_integer().
 crc32(Z, CRC, Data) when is_reference(Z) ->
     erlang:crc32(CRC, Data);
 crc32(_Z, _CRC, _Data) ->
@@ -399,10 +416,10 @@ crc32(_Z, _CRC, _Data) ->
 
 -spec crc32_combine(Z, CRC1, CRC2, Size2) -> CRC when
       Z :: zstream(),
-      CRC :: integer(),
-      CRC1 :: integer(),
-      CRC2 :: integer(),
-      Size2 :: integer().
+      CRC :: non_neg_integer(),
+      CRC1 :: non_neg_integer(),
+      CRC2 :: non_neg_integer(),
+      Size2 :: non_neg_integer().
 crc32_combine(Z, CRC1, CRC2, Size2) when is_reference(Z) ->
     erlang:crc32_combine(CRC1, CRC2, Size2);
 crc32_combine(_Z, _CRC1, _CRC2, _Size2) ->
@@ -411,7 +428,7 @@ crc32_combine(_Z, _CRC1, _CRC2, _Size2) ->
 -spec adler32(Z, Data) -> CheckSum when
       Z :: zstream(),
       Data :: iodata(),
-      CheckSum :: integer().
+      CheckSum :: non_neg_integer().
 adler32(Z, Data) when is_reference(Z) ->
     erlang:adler32(Data);
 adler32(_Z, _Data) ->
@@ -419,9 +436,9 @@ adler32(_Z, _Data) ->
 
 -spec adler32(Z, PrevAdler, Data) -> CheckSum when
       Z :: zstream(),
-      PrevAdler :: integer(),
+      PrevAdler :: non_neg_integer(),
       Data :: iodata(),
-      CheckSum :: integer().
+      CheckSum :: non_neg_integer().
 adler32(Z, Adler, Data) when is_reference(Z) ->
     erlang:adler32(Adler, Data);
 adler32(_Z, _Adler, _Data) ->
@@ -429,10 +446,10 @@ adler32(_Z, _Adler, _Data) ->
 
 -spec adler32_combine(Z, Adler1, Adler2, Size2) -> Adler when
       Z :: zstream(),
-      Adler :: integer(),
-      Adler1 :: integer(),
-      Adler2 :: integer(),
-      Size2 :: integer().
+      Adler :: non_neg_integer(),
+      Adler1 :: non_neg_integer(),
+      Adler2 :: non_neg_integer(),
+      Size2 :: non_neg_integer().
 adler32_combine(Z, Adler1, Adler2, Size2) when is_reference(Z) ->
     erlang:adler32_combine(Adler1, Adler2, Size2);
 adler32_combine(_Z, _Adler1, _Adler2, _Size2) ->
@@ -532,7 +549,7 @@ gzip(Data) ->
 gunzip(Data) ->
     Z = open(),
     Bs = try
-             inflateInit(Z, 16+?MAX_WBITS),
+             inflateInit(Z, 16+?MAX_WBITS, reset),
              B = inflate(Z, Data),
              inflateEnd(Z),
              B
@@ -652,6 +669,11 @@ arg_strategy(_) -> erlang:error(bad_compression_strategy).
 arg_method(deflated) -> ?Z_DEFLATED;
 arg_method(_) -> erlang:error(bad_compression_method).
 
+arg_eos_behavior(error) -> ?EOS_BEHAVIOR_ERROR;
+arg_eos_behavior(reset) -> ?EOS_BEHAVIOR_RESET;
+arg_eos_behavior(cut) -> ?EOS_BEHAVIOR_CUT;
+arg_eos_behavior(_) -> erlang:error(bad_eos_behavior).
+
 -spec arg_bitsz(zwindowbits()) -> zwindowbits().
 arg_bitsz(Bits) when is_integer(Bits) andalso
                      ((8 =< Bits andalso Bits < 48) orelse
@@ -678,4 +700,4 @@ enqueue_input_1(Z, IOVec) ->
     end.
 
 enqueue_nif(_Z, _IOVec) ->
-    erlang:nif_error(undef).
\ No newline at end of file
+    erlang:nif_error(undef).
